/*
 * Azande.c
 *
 * Author : Anders Zeijlon
 */ 

//==============================================================================
//--------------------------- Include Files ------------------------------------
//==============================================================================

#include "Azande.h"
#include "AzandeCrc.h"
#include <string.h>


//==============================================================================
//--------------------------- PRIVATE Types ------------------------------------
//==============================================================================


typedef union
{
	int32_t integer;
	unsigned char bytes[4];
} INT32_UNION_T;

typedef union
{
	double dValue;
	unsigned char dBuffer[4];
} DOUBLE_UNION_4_T;

typedef union
{
	double dValue;
	unsigned char dBuffer[8];
} DOUBLE_UNION_8_T;

typedef void (*AZANDE_VOID_BOOL_FUNC_PTR_T)(uint8_t);
typedef void (*AZANDE_VOID_LONG_FUNC_PTR_T)(int32_t);
typedef void (*AZANDE_VOID_DOUBLE_FUNC_PTR_T)(double);
typedef void (*AZANDE_VOID_TEXT_FUNC_PTR_T)(uint8_t*, uint16_t);

//==============================================================================
//--------------------------- PRIVATE Defines ----------------------------------
//==============================================================================

#define FIXED_SUB_ID 100

#define ID_SEND_FEATURES 			0x77
#define SUBID_COMMAND_ALL_FEATURES	0x01
#define MSG_ID_COMMAND_ALL_FEATURES	0x77, 0x01
#define MSG_ID_ONE_FEATURE			0x77, 0x02
#define MSG_ID_TARGET_INFO			0x77, 0x03
#define MSG_ID_FEATURE_LIST_CHANGED	0x77, 0x04

#define MSG_START 			0xAA
#define MSG_END 			0xBB
#define MSG_ESCAPE 			0xCC
#define MSG_ESCAPE_START 	0xDD
#define MSG_ESCAPE_END 		0xEE
#define MSG_ESCAPE_ESCAPE 	0xFF

#define CRC_SEED_VALUE 0xFFFF

#ifdef MCU_IS_BIG_ENDIAN
	#define CRC_LOW		0
	#define CRC_HIGH	1
#else
	#define CRC_LOW		1
	#define CRC_HIGH	0
#endif

#define true 1
#define false 0

//==============================================================================
//--------------------------- PRIVATE Const --------------------------------
//==============================================================================

static const __FLASH_DIRECTIVE char msgNoPayload[1] = {0x00};


//==============================================================================
//--------------------------- PRIVATE Variables --------------------------------
//==============================================================================
static uint8_t rawInBuffer[READ_BUFFER_SIZE];
static uint8_t tempBuffer[8];
static uint16_t rawInBufferIndex;
static uint16_t unescapedInIndex;
static uint16_t sendCrc;
static uint8_t isStartByteAccepted;
static AZANDE_COMMAND const __FLASH_DIRECTIVE * command_list[AZANDE_MAX_COMMANDS];
static uint16_t command_count = 0;
static const __FLASH_DIRECTIVE char* targetInfo = 0;
static char const __FLASH_DIRECTIVE * event_xml_list[AZANDE_MAX_EVENTS];
static uint16_t event_xml_count = 0;


//==============================================================================
//--------------------------- PRIVATE (Function Prototypes) --------------------
//==============================================================================

static void evaluateMessage(void);
static uint8_t unescapeBuffer(void);
static uint8_t isCrcOk(void);
static void exeMatchingCommand(void);
static void writeWithEscape(uint8_t data);
static void blockWriteWithEscape(uint8_t data);
static void sendAllFeatures(void);
static void sendOneStaticMessage(const __FLASH_DIRECTIVE char* data, uint8_t id, uint8_t subId);
static void convertSingle2DoublePrecisionArray(uint8_t* destination_array, const double value);
static DOUBLE_UNION_4_T convertDoublePrecisionArray2Single(uint8_t* big_endian_source_array);
static double getDoubleValue(uint8_t* big_endian_source_array);
static int32_t getIntValue(uint8_t* big_endian_source_array);


//==============================================================================
//--------------------------- Public Functions ---------------------------------
//==============================================================================

void Azande__Init(const __FLASH_DIRECTIVE char* text)
{
	targetInfo = text;
}

void Azande__ReadStream(void)
{
	uint8_t data;


	while (SerialAnyBytesToRead())
	{
		if (SerialReadNextByte(data) == 1)
		{
			if (data == MSG_START)
			{
				rawInBufferIndex = 0;
				isStartByteAccepted = true;
			}
			else if (data == MSG_END)
			{
				if (isStartByteAccepted)
				{
					evaluateMessage();
					isStartByteAccepted = false;
					return;
				}
				isStartByteAccepted = false;
			}
			else
			{
				rawInBuffer[rawInBufferIndex] = data;

				++rawInBufferIndex;
				if (rawInBufferIndex >= READ_BUFFER_SIZE) // the message was too long
				{
					rawInBufferIndex = 0;
					isStartByteAccepted = false;
				}
			}

		}
	}

}

void Azande__SendEventBool(const __FLASH_DIRECTIVE AZANDE_EVENT* event_ptr, uint8_t isTrue)
{
	uint8_t crc_msb;
	uint8_t crc_lsb;

	WriteOneByteQueued(0xAA);

	sendCrc = CRC_SEED_VALUE;

	// Message Id
	writeWithEscape(event_ptr->id);
	writeWithEscape(FIXED_SUB_ID);

	// Payload
	if (isTrue > 0)
	{
		writeWithEscape(true);
	}
	else
	{
		writeWithEscape(false);
	}

	// CRC
	crc_msb = sendCrc / 256;
	crc_lsb = sendCrc % 256;

	writeWithEscape(crc_msb);
	writeWithEscape(crc_lsb);

	// End byte
	WriteOneByteQueued(0xBB);
}

void Azande__SendEventInt32(const __FLASH_DIRECTIVE AZANDE_EVENT* event_ptr, int32_t value)
{
	uint8_t crc_msb;
	uint8_t crc_lsb;
	INT32_UNION_T unionen;

	unionen.integer = value;

	// Start byte
	WriteOneByteQueued(0xAA);

	sendCrc = CRC_SEED_VALUE;

	// Message Id
	writeWithEscape(event_ptr->id);
	writeWithEscape(FIXED_SUB_ID);

	// Payload
	// The value is sent as little endian
#ifdef MCU_IS_BIG_ENDIAN
	// Big => Big
	writeWithEscape(unionen.bytes[0]);
	writeWithEscape(unionen.bytes[1]);
	writeWithEscape(unionen.bytes[2]);
	writeWithEscape(unionen.bytes[3]);
#else
	// Little => Big (reverse order)
	writeWithEscape(unionen.bytes[3]);
	writeWithEscape(unionen.bytes[2]);
	writeWithEscape(unionen.bytes[1]);
	writeWithEscape(unionen.bytes[0]);
#endif

	// CRC
	crc_msb = sendCrc / 256;
	crc_lsb = sendCrc % 256;

	writeWithEscape(crc_msb);
	writeWithEscape(crc_lsb);

	// End byte
	WriteOneByteQueued(0xBB);
}

void Azande__SendEventText(const __FLASH_DIRECTIVE AZANDE_EVENT* event_ptr, char* text)
{
	uint8_t crc_msb;
	uint8_t crc_lsb;
	uint16_t i;

	// Start byte
	WriteOneByteQueued(0xAA);

	sendCrc = CRC_SEED_VALUE;

	// Message Id
	writeWithEscape(event_ptr->id);
	writeWithEscape(FIXED_SUB_ID);

	// Payload
	for (i = 0; text[i] != 0; i++)
	{
		writeWithEscape(text[i]);
	}

	// CRC
	crc_msb = sendCrc / 256;
	crc_lsb = sendCrc % 256;

	writeWithEscape(crc_msb);
	writeWithEscape(crc_lsb);

	// End byte
	WriteOneByteQueued(0xBB);
}

void Azande__SendEventDouble(const __FLASH_DIRECTIVE AZANDE_EVENT* event_ptr, double value)
{
	int8_t i;
	uint8_t crc_msb;
	uint8_t crc_lsb;

	// Start byte
	WriteOneByteQueued(0xAA);

	sendCrc = CRC_SEED_VALUE;

	// Message Id
	writeWithEscape(event_ptr->id);
	writeWithEscape(FIXED_SUB_ID);

	// Payload
	if (sizeof(double) == 8) // Double precision (8-bytes)
	{
		memcpy(tempBuffer, &value, 8);
	}
	else if (sizeof(double) == 4) // Single precision (4-bytes)
	{
		convertSingle2DoublePrecisionArray(tempBuffer, value);
	}
	else
	{
		// Not supported
		for (i = 0; i < 8; i++)
		{
			tempBuffer[i] = 0;
		}
	}

#ifdef MCU_IS_LITTLE_ENDIAN
	// Little => Big (reverse order)
	for (i = 7; i >= 0; i--)
	{
		writeWithEscape(tempBuffer[i]);
	}
#else
	// Big => Big (keep order)
	for (i = 0; i < 8; i++)
	{
		writeWithEscape(tempBuffer[i]);
	}
#endif

	// CRC
	crc_msb = sendCrc / 256;
	crc_lsb = sendCrc % 256;

	writeWithEscape(crc_msb);
	writeWithEscape(crc_lsb);

	// End byte
	WriteOneByteQueued(0xBB);
}

void Azande__AddEvent(const __FLASH_DIRECTIVE AZANDE_EVENT* event_ptr)
{
	event_xml_list[event_xml_count] = event_ptr->xml;
	++event_xml_count;
}

void Azande__AddCommand(const __FLASH_DIRECTIVE AZANDE_COMMAND* cmd_ptr)
{
	command_list[command_count] = cmd_ptr;
	++command_count;
}

void Azande__Clear()
{
	command_count = 0;
	event_xml_count = 0;

}

void Azande__Notify()
{
	sendOneStaticMessage(msgNoPayload, MSG_ID_FEATURE_LIST_CHANGED);
}


//==============================================================================
//--------------------------- Private Functions ---------------------------------
//==============================================================================



static void evaluateMessage(void)
{
	if (rawInBufferIndex > 3)
	{
		if (unescapeBuffer() == true)
		{
			if (isCrcOk() == true)
			{
				exeMatchingCommand();
			}
		}
	}
}

static void sendAllFeatures(void)
{
	uint16_t i;

	StartBlockWriting();

	if (targetInfo != 0)
	{
		sendOneStaticMessage(targetInfo, MSG_ID_TARGET_INFO);
	}

	for (i = 0; i < command_count; i++)
	{
		sendOneStaticMessage(command_list[i]->xml, MSG_ID_ONE_FEATURE);
	}

	for (i = 0; i < event_xml_count; i++)
	{
		sendOneStaticMessage(event_xml_list[i], MSG_ID_ONE_FEATURE);
	}

	StopBlockWriting();
}

static void sendOneStaticMessage(const __FLASH_DIRECTIVE char* data, uint8_t id, uint8_t subId)
{
	uint8_t crc_msb;
	uint8_t crc_lsb;


	// Start byte
	WriteOneByteBlocked(0xAA);

	sendCrc = CRC_SEED_VALUE;

	// Message Id
	blockWriteWithEscape(id);

	// Message Sub Id
	blockWriteWithEscape(subId);

	// Payload
	while (*data != 0)
	{
		blockWriteWithEscape(*data);
		++data;
	}

	// CRC
	crc_msb = sendCrc / 256;
	crc_lsb = sendCrc % 256;

	blockWriteWithEscape(crc_msb);
	blockWriteWithEscape(crc_lsb);

	// End byte
	WriteOneByteBlocked(0xBB);

}

static uint8_t unescapeBuffer(void)
{
	uint8_t escapeFound = false;
	uint16_t i;

	unescapedInIndex = 0;

	for (i = 0; i < rawInBufferIndex; i++)
	{
		if (escapeFound == true)
		{
			escapeFound = false;

			if (rawInBuffer[i] == MSG_ESCAPE_START)
			{
				rawInBuffer[unescapedInIndex] = MSG_START;
				++unescapedInIndex;
			}
			else if (rawInBuffer[i] == MSG_ESCAPE_END)
			{
				rawInBuffer[unescapedInIndex] = MSG_END;
				++unescapedInIndex;
			}
			else if (rawInBuffer[i] == MSG_ESCAPE_ESCAPE)
			{
				rawInBuffer[unescapedInIndex] = MSG_ESCAPE;
				++unescapedInIndex;
			}
			else
			{
				// Incorrect format found. Escape followed of a byte not MSG_ESCAPE_START, MSG_ESCAPE_END or MSG_ESCAPE_ESCAPE
				unescapedInIndex = 0;
				return false;
			}
		}
		else if (rawInBuffer[i] == MSG_ESCAPE)
		{
			escapeFound = true;
		}
		else
		{
			rawInBuffer[unescapedInIndex] = rawInBuffer[i];
			++unescapedInIndex;
		}
	}
	return true;
}

static uint8_t isCrcOk(void)
{
	uint16_t i;
	uint16_t crc = CRC_SEED_VALUE;

	for (i = 0; i < (unescapedInIndex - 2); i++)
	{
		crc = Crc_Update(crc, rawInBuffer[i]);
	}

	if ((rawInBuffer[unescapedInIndex - 1] == (crc % 256)) && (rawInBuffer[unescapedInIndex - 2] == (crc / 256)))
	{
		return true;
	}
	else
	{
		return false;
	}
}

static void exeMatchingCommand(void)
{
	uint16_t i;
	uint8_t id;
	uint8_t subId;

	id = rawInBuffer[0];
	subId = rawInBuffer[1];

	if ((id == ID_SEND_FEATURES) && (subId == SUBID_COMMAND_ALL_FEATURES))
	{
		sendAllFeatures();
	}
	else
	{
		for (i = 0; i < command_count; i++)
		{
			if ((id == command_list[i]->id) && (subId == FIXED_SUB_ID))
			{
				switch (command_list[i]->type)
				{
				case AZANDE_COMMAND_TYPE_VOID:
					((AZANDE_VOID_VOID_FUNC_PTR_T)(command_list[i]->cmd_function))();
					break;

				case AZANDE_COMMAND_TYPE_BOOL:
					((AZANDE_VOID_BOOL_FUNC_PTR_T)(command_list[i]->cmd_function))(rawInBuffer[2] == 1);
					break;

				case AZANDE_COMMAND_TYPE_ENUM:
				case AZANDE_COMMAND_TYPE_INT:
					((AZANDE_VOID_LONG_FUNC_PTR_T)(command_list[i]->cmd_function))(getIntValue(&rawInBuffer[2]));
					break;

				case AZANDE_COMMAND_TYPE_DOUBLE:
					((AZANDE_VOID_DOUBLE_FUNC_PTR_T)(command_list[i]->cmd_function))(getDoubleValue(&rawInBuffer[2]));
					break;

				case AZANDE_COMMAND_TYPE_TEXT:
					((AZANDE_VOID_TEXT_FUNC_PTR_T)(command_list[i]->cmd_function))(&(rawInBuffer[2]), unescapedInIndex - 4);
					break;

				}
			}
		}
	}
}

static void writeWithEscape(uint8_t data)
{
	uint8_t new_data;

	new_data = data;

	sendCrc = Crc_Update(sendCrc, data);

	if (data == MSG_START)
	{
		WriteOneByteQueued(MSG_ESCAPE);
		new_data = MSG_ESCAPE_START;
	}
	else if (data == MSG_END)
	{
		WriteOneByteQueued(MSG_ESCAPE);
		new_data = MSG_ESCAPE_END;
	}
	else if (data == MSG_ESCAPE)
	{
		WriteOneByteQueued(MSG_ESCAPE);
		new_data = MSG_ESCAPE_ESCAPE;
	}

	WriteOneByteQueued(new_data);
}

static void blockWriteWithEscape(uint8_t data)
{
	uint8_t new_data;

	new_data = data;

	sendCrc = Crc_Update(sendCrc, data);

	if (data == MSG_START)
	{
		WriteOneByteBlocked(MSG_ESCAPE);
		new_data = MSG_ESCAPE_START;
	}
	else if (data == MSG_END)
	{
		WriteOneByteBlocked(MSG_ESCAPE);
		new_data = MSG_ESCAPE_END;
	}
	else if (data == MSG_ESCAPE)
	{
		WriteOneByteBlocked(MSG_ESCAPE);
		new_data = MSG_ESCAPE_ESCAPE;
	}

	WriteOneByteBlocked(new_data);

}

static void convertSingle2DoublePrecisionArray(uint8_t* destination_array, const double value)
{
	volatile DOUBLE_UNION_4_T Double_Union_4_Value;
	Double_Union_4_Value.dValue = value;

#ifdef MCU_IS_LITTLE_ENDIAN
	// Convert and keep as little endian
	destination_array[7] = ((Double_Union_4_Value.dBuffer[3]  >> 3) & 0x07); 	// bit 27-29 ==>> bit 56-58
	destination_array[7] |= (Double_Union_4_Value.dBuffer[3] & 0xC0); 			// bit 30-31 ==>> bit 62-63
	if ((Double_Union_4_Value.dValue < 2) && (Double_Union_4_Value.dValue > -2))
	{
		destination_array[7] |= 0x38; 											// set bit 59-60
	}
	destination_array[6] = (Double_Union_4_Value.dBuffer[3] << 5);				// bit 24-29 ==>> bit 53-55
	destination_array[6] |= (Double_Union_4_Value.dBuffer[2] >> 3);				// bit 19-23 ==>> bit 48-52
	destination_array[5] = (Double_Union_4_Value.dBuffer[2] << 5);				// bit 16-18 ==>> bit 45-47
	destination_array[5] |= (Double_Union_4_Value.dBuffer[1] >> 3);				// bit 11-15 ==>> bit 40-44
	destination_array[4] = (Double_Union_4_Value.dBuffer[1] << 5);				// bit 8 -10 ==>> bit 37-39
	destination_array[4] |= (Double_Union_4_Value.dBuffer[0] >> 3);				// bit 3 - 7 ==>> bit 32-36
	destination_array[3] = (Double_Union_4_Value.dBuffer[0] << 5);				// bit 0 - 2 ==>> bit 29-31
	destination_array[2] = 0;													// clear
	destination_array[1] = 0;													// clear
	destination_array[0] = 0;													// clear
#else // Big Endian
	// Convert and keep as big endian
	destination_array[0] = ((Double_Union_4_Value.dBuffer[0]  >> 3) & 0x07); // MSB
	destination_array[0] |= (Double_Union_4_Value.dBuffer[0] & 0xC0);
	if ((Double_Union_4_Value.dValue < 2) && (Double_Union_4_Value.dValue > -2))
	{
		destination_array[0] |= 0x38; 											// set bit 59-60
	}
	destination_array[1] = (Double_Union_4_Value.dBuffer[0] << 5);
	destination_array[1] += (Double_Union_4_Value.dBuffer[1] >> 3);
	destination_array[2] = (Double_Union_4_Value.dBuffer[1] << 5);
	destination_array[2] += (Double_Union_4_Value.dBuffer[2] >> 3);
	destination_array[3] = (Double_Union_4_Value.dBuffer[2] << 5);
	destination_array[3] += (Double_Union_4_Value.dBuffer[3] >> 3);
	destination_array[4] = (Double_Union_4_Value.dBuffer[3] << 5);
	destination_array[5] = 0;
	destination_array[6] = 0;
	destination_array[7] = 0;											// LSB
#endif

}

static DOUBLE_UNION_4_T convertDoublePrecisionArray2Single(uint8_t* big_endian_source_array)
{
	volatile DOUBLE_UNION_4_T Double_Union_4_Value;

#ifdef MCU_IS_LITTLE_ENDIAN
	// Little => Big (reverse order)
	Double_Union_4_Value.dBuffer[3] = (big_endian_source_array[1] >> 5); // MSB
	Double_Union_4_Value.dBuffer[3] |= ((big_endian_source_array[0] & 0x07 ) << 3); // MSB
	Double_Union_4_Value.dBuffer[3] |= (big_endian_source_array[0] & 0xC0); // MSB

	Double_Union_4_Value.dBuffer[2] = (big_endian_source_array[2] >> 5);
	Double_Union_4_Value.dBuffer[2] += (big_endian_source_array[1] << 3);

	Double_Union_4_Value.dBuffer[1] = (big_endian_source_array[3] >> 5);
	Double_Union_4_Value.dBuffer[1] += (big_endian_source_array[2] << 3);

	Double_Union_4_Value.dBuffer[0] = (big_endian_source_array[4] >> 5); // LSB
	Double_Union_4_Value.dBuffer[0] += (big_endian_source_array[3] << 3); // LSB
#else
	// Big => Big (keep order)
	Double_Union_4_Value.dBuffer[0] = (big_endian_source_array[1] >> 5); // MSB
	Double_Union_4_Value.dBuffer[0] |= ((big_endian_source_array[0] & 0x07 ) << 3); // MSB
	Double_Union_4_Value.dBuffer[0] |= (big_endian_source_array[0] & 0xC0); // MSB

	Double_Union_4_Value.dBuffer[1] = (big_endian_source_array[2] >> 5);
	Double_Union_4_Value.dBuffer[1] += (big_endian_source_array[1] << 3);

	Double_Union_4_Value.dBuffer[2] = (big_endian_source_array[3] >> 5);
	Double_Union_4_Value.dBuffer[2] += (big_endian_source_array[2] << 3);

	Double_Union_4_Value.dBuffer[3] = (big_endian_source_array[4] >> 5); // LSB
	Double_Union_4_Value.dBuffer[3] += (big_endian_source_array[3] << 3); // LSB
#endif

	return Double_Union_4_Value;
}

static double getDoubleValue(uint8_t* big_endian_source_array)
{
	uint8_t i;
	volatile DOUBLE_UNION_8_T Double_Union_8_Value;

	if (sizeof(double) == 8)
	{
#ifdef MCU_IS_LITTLE_ENDIAN
		// Little => Big (reverse order)
		for (i = 0; i < 8; i++)
		{
			Double_Union_8_Value.dBuffer[i] = big_endian_source_array[7 - i];
		}
#else
		// Big => Big (keep order)
		for (i = 0; i < 8; i++)
		{
			Double_Union_8_Value.dBuffer[i] = big_endian_source_array[i];
		}
#endif
		return Double_Union_8_Value.dValue;
	}
	else if (sizeof(double) == 4)
	{
		return convertDoublePrecisionArray2Single(big_endian_source_array).dValue;
	}
	else
	{
		// Not supported
		return 0.0;
	}
}

static int32_t getIntValue(uint8_t* big_endian_source_array)
{
	uint8_t i;
	volatile INT32_UNION_T int_union;

#ifdef MCU_IS_LITTLE_ENDIAN
		// Little => Big (reverse order)
		for (i = 0; i < 4; i++)
		{
			int_union.bytes[i] = big_endian_source_array[3 - i];
		}
#else
		// Big => Big (keep order)
		for (i = 0; i < 4; i++)
		{
			int_union.bytes[i] = big_endian_source_array[i];
		}
#endif
		return int_union.integer;

}



