/*
 * PcUart.c
 *
 * Author : Anders Zeijlon
 */ 


//==============================================================================
//--------------------------- Include Files ------------------------------------
//==============================================================================

#include "PcUart.h"

#include <avr/interrupt.h>
#include <stdint.h>



//==============================================================================
//--------------------------- PRIVATE (Variables, Constants & Defines) ---------
//==============================================================================
#define RX_BUFFER_SIZE 50
#define TX_BUFFER_SIZE 50


static volatile uint8_t rx_buffer[RX_BUFFER_SIZE];
static volatile uint8_t rx_write_index = 0;
static volatile uint8_t rx_read_index = 0;

static volatile uint8_t tx_buffer[TX_BUFFER_SIZE];
static volatile uint8_t tx_write_index = 0;
static volatile uint8_t tx_read_index = 0;

static volatile uint8_t isBlockWriting;

//==============================================================================
//--------------------------- PRIVATE (Function Prototypes) --------------------
//==============================================================================

static void usart_tx_send_next_byte(void);



//==============================================================================
//--------------------------- Public Functions ---------------------------------
//==============================================================================

void PcUart_init(void)
{
	rx_write_index = 0;
	rx_read_index = 0;

	tx_write_index = 0;
	tx_read_index = 0;
	
	isBlockWriting = 0;

	cli(); // Disable interrupts globally.


	UCSR0A = 0;
	UCSR0B = 0;
	UCSR0C = 0;


	
	//UBRR0 = 51; // Baudrate = 9 600  error 0.2%
	UBRR0 = 25; // Baudrate = 19 200  error 0.2%
	//UBRR0 = 12; // Baudrate = 38 400 error 0.2%
	//UBRR0 = 8; // Baudrate = 57 600 error -3.5%
	//UBRR0 = 3; // Baudrate = 115 200 error 8.5%
	//UBRR0 = 1; // Baudrate = 230 400 error 8.5%
	//UBRR0 = 0; // Baudrate = 460 800 error 8.5%

	UCSR0C = 0x06; // 00 00 0 11 0 => 0x06 Asynchronous USART, No Parity, 1 stop-bit, 8-bit characters

	UCSR0B = 0xB8; // 1011 1000 => 0xB8 RX Complete Interrupt Enable, USART Data Register Empty Interrupt Enable,  Receive Enable, Transmitter Enable

	sei(); // Enable interrupts globally.

}

// Returns how many bytes read
uint16_t PcUart_Read(uint8_t* destination, uint16_t max_count)
{
	uint8_t i;

	for (i = 0; i < max_count; i++)
	{
		if (rx_read_index != rx_write_index)
		{
			destination[i] =  rx_buffer[rx_read_index];
			rx_read_index++;
			if (rx_read_index >= RX_BUFFER_SIZE)
			{
				rx_read_index = 0;
			}
		}
		else
		{
			return i;
		}
	}

	return i;
}


uint8_t PcUart_AnyToRead(void)
{
	if (rx_read_index != rx_write_index)
	{
		return 1;
	}
	else
	{
		return 0;
	}
}


void PcUart_WriteOneByte(uint8_t byte)
{
	uint8_t data = byte;
	PcUart_Write(&data, 1);
}



void PcUart_Write(uint8_t* source, uint16_t count)
{
	uint16_t i;

	for (i = 0; i < count; i++)
	{
		tx_buffer[tx_write_index] =  source[i];

		tx_write_index++;
		if (tx_write_index >= TX_BUFFER_SIZE)
		{
			tx_write_index = 0;
		}
	}
	
	if (UCSR0A & 0x20) // USART Data Register Empty
	{
		usart_tx_send_next_byte();
	}
}

void PcUart_StartBlockWriting(void)
{
	cli();
	UCSR0B = 0x18; // 0001 1000 => 0xB8 Receive Enable, Transmitter Enable
	isBlockWriting = 1;
}

void PcUart_StopBlockWriting(void)
{
	PcUart_init();
}

void PcUart_BlockWriteOneByte(uint8_t byte)
{
	while ((UCSR0A & 0x20) == 0); // USART Data Register Empty
	UCSR0A &= 0x40; // Clear TXCn: USART Transmit Complete
	UDR0 = byte;
	while ((UCSR0A & 0x40) == 0); // TXCn: USART Transmit Complete
}


//==============================================================================
//--------------------------- Private Functions --------------------------------
//==============================================================================

ISR(USART_RX_vect)
{
	uint8_t is_com_error = 0;
	uint8_t data_byte = 0;

	data_byte = UDR0; // Should also clear "USART Receive Complete" flag

	is_com_error = (UCSR0A & 0x18); // bit4 = Frame Error, bit3 = Data OverRun
	UCSR0A = 0;

	if (is_com_error == 0)
	{
		// No error
		rx_buffer[rx_write_index] = data_byte;

		rx_write_index++;
		if (rx_write_index >= RX_BUFFER_SIZE)
		{
			rx_write_index = 0;
		}
	}
}




static void usart_tx_send_next_byte(void)
{
	UCSR0B &= 0xDF; // Data Register Empty Interrupt Disable

	if ((UCSR0A & 0x20) && (tx_read_index != tx_write_index))
	{
		UDR0 = tx_buffer[tx_read_index]; // Load new value => Send

		tx_read_index++;
		if (tx_read_index >= TX_BUFFER_SIZE)
		{
			tx_read_index = 0;
		}

		UCSR0B |= 0x20; // Data Register Empty Interrupt Enable
	}
}


ISR(USART_UDRE_vect)
{
	if ((isBlockWriting == 0) && (UCSR0A & 0x20)) // USART Data Register Empty
	{
		usart_tx_send_next_byte();
	}
}
